@@ -1,20 +1,16 @@
 # coding=utf-8
-"""
-url.py - Sopel URL title module
-Copyright 2010-2011, Michael Yanovich, yanovich.net, Kenneth Sham
-Copyright 2012-2013 Edward Powell
-Copyright 2013      Lior Ramati (firerogue517@gmail.com)
-Copyright © 2014 Elad Alfassa <elad@fedoraproject.org>
-Licensed under the Eiffel Forum License 2.
-
-http://sopel.chat
-"""
+"""URL title module"""
+# Copyright 2010-2011, Michael Yanovich, yanovich.net, Kenneth Sham
+# Copyright 2012-2013 Elsie Powell
+# Copyright 2013      Lior Ramati (firerogue517@gmail.com)
+# Copyright © 2014 Elad Alfassa <elad@fedoraproject.org>
+# Licensed under the Eiffel Forum License 2.
 from __future__ import unicode_literals, absolute_import, print_function, division
 
 import re
 from sopel import web, tools
 from sopel.module import commands, rule, example
-from sopel.config.types import ValidatedAttribute, StaticSection
+from sopel.config.types import ValidatedAttribute, ListAttribute, StaticSection
 
 
 url_finder = None
@@ -33,7 +29,7 @@ max_bytes = 655360
 
 class UrlSection(StaticSection):
     # TODO some validation rules maybe?
-    exclude = ValidatedAttribute('exclude')
+    exclude = ListAttribute('exclude')
     exclusion_char = ValidatedAttribute('exclusion_char', default='!')
 
 
